@@ -1,5 +1,5 @@
 /* POSIX-based operating system interface for GNU Make.
-Copyright (C) 2016-2022 Free Software Foundation, Inc.
+Copyright (C) 2016-2023 Free Software Foundation, Inc.
 This file is part of GNU Make.
 
 GNU Make is free software; you can redistribute it and/or modify it under the
@@ -12,7 +12,7 @@ WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License along with
-this program.  If not, see <http://www.gnu.org/licenses/>.  */
+this program.  If not, see <https://www.gnu.org/licenses/>.  */
 
 #include "makeint.h"
 
@@ -24,6 +24,10 @@ this program.  If not, see <http://www.gnu.org/licenses/>.  */
 #elif defined(HAVE_SYS_FILE_H)
 # include <sys/file.h>
 #endif
+#if MK_OS_ZOS
+/* FIXME: HAVE_PSELECT path hangs on z/OS */
+#undef HAVE_PSELECT
+#endif
 
 #if !defined(FD_OK)
 # define FD_OK(_f) 1
@@ -75,6 +79,9 @@ check_io_state ()
 
 /* This section provides OS-specific functions to support the jobserver.  */
 
+/* True if this is the root make instance.  */
+static unsigned char job_root = 0;
+
 /* These track the state of the jobserver pipe.  Passed to child instances.  */
 static int job_fds[2] = { -1, -1 };
 
@@ -129,6 +136,9 @@ set_blocking (int fd, int blocking)
       if (r < 0)
         pfatal_with_name ("fcntl(O_NONBLOCK)");
     }
+#else
+  (void) fd;
+  (void) blocking;
 #endif
 }
 
@@ -140,7 +150,16 @@ jobserver_setup (int slots, const char *style)
 #if HAVE_MKFIFO
   if (style == NULL || strcmp (style, "fifo") == 0)
     {
-      fifo_name = get_tmppath ();
+  /* Unfortunately glibc warns about uses of mktemp even though we aren't
+     using it in dangerous way here.  So avoid this by generating our own
+     temporary file name.  */
+# define  FNAME_PREFIX "GMfifo"
+      const char *tmpdir = get_tmpdir ();
+
+      fifo_name = xmalloc (strlen (tmpdir) + CSTRLEN (FNAME_PREFIX)
+                           + INTSTR_LENGTH + 2);
+      sprintf (fifo_name, "%s/" FNAME_PREFIX "%" MK_PRI64_PREFIX "d",
+               tmpdir, (long long)make_pid ());
 
       EINTRLOOP (r, mkfifo (fifo_name, 0600));
       if (r < 0)
@@ -155,16 +174,14 @@ jobserver_setup (int slots, const char *style)
              hang until the write side is open.  */
           EINTRLOOP (job_fds[0], open (fifo_name, O_RDONLY|O_NONBLOCK));
           if (job_fds[0] < 0)
-            OSS (fatal, NILF, _("Cannot open jobserver %s: %s"),
+            OSS (fatal, NILF, _("cannot open jobserver %s: %s"),
                  fifo_name, strerror (errno));
 
           EINTRLOOP (job_fds[1], open (fifo_name, O_WRONLY));
           if (job_fds[0] < 0)
-            OSS (fatal, NILF, _("Cannot open jobserver %s: %s"),
+            OSS (fatal, NILF, _("cannot open jobserver %s: %s"),
                  fifo_name, strerror (errno));
 
-          DB (DB_JOBS, (_("Jobserver setup (fifo %s)\n"), fifo_name));
-
           js_type = js_fifo;
         }
     }
@@ -173,15 +190,12 @@ jobserver_setup (int slots, const char *style)
   if (js_type == js_none)
     {
       if (style && strcmp (style, "pipe") != 0)
-        OS (fatal, NILF, _("Unknown jobserver auth style '%s'"), style);
+        OS (fatal, NILF, _("unknown jobserver auth style '%s'"), style);
 
       EINTRLOOP (r, pipe (job_fds));
       if (r < 0)
         pfatal_with_name (_("creating jobs pipe"));
 
-      DB (DB_JOBS, (_("Jobserver setup (fds %d,%d)\n"),
-                    job_fds[0], job_fds[1]));
-
       js_type = js_pipe;
     }
 
@@ -203,6 +217,8 @@ jobserver_setup (int slots, const char *style)
   /* When using pselect() we want the read to be non-blocking.  */
   set_blocking (job_fds[0], 0);
 
+  job_root = 1;
+
   return 1;
 }
 
@@ -220,21 +236,24 @@ jobserver_parse_auth (const char *auth)
 
       EINTRLOOP (job_fds[0], open (fifo_name, O_RDONLY));
       if (job_fds[0] < 0)
-        OSS (fatal, NILF,
-             _("Cannot open jobserver %s: %s"), fifo_name, strerror (errno));
+        {
+          OSS (error, NILF,
+               _("cannot open jobserver %s: %s"), fifo_name, strerror (errno));
+          return 0;
+        }
 
       EINTRLOOP (job_fds[1], open (fifo_name, O_WRONLY));
-      if (job_fds[0] < 0)
-        OSS (fatal, NILF,
-             _("Cannot open jobserver %s: %s"), fifo_name, strerror (errno));
-
+      if (job_fds[1] < 0)
+        {
+          OSS (error, NILF,
+               _("cannot open jobserver %s: %s"), fifo_name, strerror (errno));
+          return 0;
+        }
       js_type = js_fifo;
     }
   /* If not, it must be a simple pipe.  */
   else if (sscanf (auth, "%d,%d", &rfd, &wfd) == 2)
     {
-      DB (DB_JOBS, (_("Jobserver client (fds %d,%d)\n"), rfd, wfd));
-
       /* The parent overrode our FDs because we aren't a recursive make.  */
       if (rfd == -2 || wfd == -2)
         return 0;
@@ -327,8 +346,20 @@ jobserver_clear ()
 
   job_fds[0] = job_fds[1] = job_rfd = -1;
 
-  free (fifo_name);
-  fifo_name = NULL;
+  if (fifo_name)
+    {
+      if (job_root)
+        {
+          int r;
+          EINTRLOOP (r, unlink (fifo_name));
+        }
+
+      if (!handling_fatal_signal)
+        {
+          free (fifo_name);
+          fifo_name = NULL;
+        }
+    }
 
   js_type = js_none;
 }
@@ -368,11 +399,10 @@ jobserver_acquire_all ()
       ++tokens;
     }
 
-  if (fifo_name)
-    EINTRLOOP (r, unlink (fifo_name));
-
   DB (DB_JOBS, ("Acquired all %u jobserver tokens.\n", tokens));
 
+  jobserver_clear ();
+
   return tokens;
 }
 
@@ -462,7 +492,7 @@ jobserver_acquire (int timeout)
           case EBADF:
             /* Someone closed the jobs pipe.
                That shouldn't happen but if it does we're done.  */
-              O (fatal, NILF, _("job server shut down"));
+            O (fatal, NILF, _("job server shut down"));
 
           default:
             pfatal_with_name (_("pselect jobs pipe"));
@@ -594,7 +624,7 @@ jobserver_acquire (int timeout)
      go back and reap_children(), and try again.  */
   errno = saved_errno;
 
-  if (errno != EINTR && errno != EBADF)
+  if (errno != EINTR && errno != EBADF && errno != EAGAIN)
     pfatal_with_name (_("read jobs pipe"));
 
   if (errno == EBADF)
@@ -615,7 +645,7 @@ static int osync_handle = -1;
 
 static char *osync_tmpfile = NULL;
 
-static unsigned int sync_parent = 0;
+static unsigned int sync_root = 0;
 
 unsigned int
 osync_enabled ()
@@ -627,8 +657,8 @@ void
 osync_setup ()
 {
   osync_handle = get_tmpfd (&osync_tmpfile);
-  if (osync_handle >= 0)
-    sync_parent = 1;
+  fd_noinherit (osync_handle);
+  sync_root = 1;
 }
 
 char *
@@ -655,6 +685,7 @@ osync_parse_mutex (const char *mutex)
       return 0;
     }
 
+  free (osync_tmpfile);
   osync_tmpfile = xstrdup (mutex + CSTRLEN (MUTEX_PREFIX));
 
   EINTRLOOP (osync_handle, open (osync_tmpfile, O_WRONLY));
@@ -662,21 +693,26 @@ osync_parse_mutex (const char *mutex)
     OSS (fatal, NILF, _("cannot open output sync mutex %s: %s"),
          osync_tmpfile, strerror (errno));
 
+  fd_noinherit (osync_handle);
+
   return 1;
 }
 
 void
 osync_clear ()
 {
-  if (osync_handle)
+  if (osync_handle >= 0)
     {
       close (osync_handle);
       osync_handle = -1;
     }
 
-  if (sync_parent && osync_tmpfile)
+  if (sync_root && osync_tmpfile)
     {
-      unlink (osync_tmpfile);
+      int r;
+
+      EINTRLOOP (r, unlink (osync_tmpfile));
+      free (osync_tmpfile);
       osync_tmpfile = NULL;
     }
 }
@@ -692,6 +728,7 @@ osync_acquire ()
       fl.l_whence = SEEK_SET;
       fl.l_start = 0;
       fl.l_len = 1;
+      /* We don't want to keep waiting on EINTR.  */
       if (fcntl (osync_handle, F_SETLKW, &fl) == -1)
         {
           perror ("fcntl()");
@@ -713,6 +750,7 @@ osync_release ()
       fl.l_whence = SEEK_SET;
       fl.l_start = 0;
       fl.l_len = 1;
+      /* We don't want to keep waiting on EINTR.  */
       if (fcntl (osync_handle, F_SETLKW, &fl) == -1)
         perror ("fcntl()");
     }
@@ -736,7 +774,7 @@ get_bad_stdin ()
       if (pipe (pd) == 0)
         {
           /* Close the write side.  */
-          (void) close (pd[1]);
+          close (pd[1]);
           /* Save the read side.  */
           bad_stdin = pd[0];
 
@@ -809,3 +847,50 @@ fd_set_append (int fd)
     }
 #endif
 }
+
+/* Return a file descriptor for a new anonymous temp file, or -1.  */
+int
+os_anontmp ()
+{
+  const char *tdir = get_tmpdir ();
+  int fd = -1;
+
+#ifdef O_TMPFILE
+  static unsigned int tmpfile_works = 1;
+
+  if (tmpfile_works)
+    {
+      EINTRLOOP (fd, open (tdir, O_RDWR | O_TMPFILE | O_EXCL, 0600));
+      if (fd >= 0)
+        return fd;
+
+      DB (DB_BASIC, (_("Cannot open '%s' with O_TMPFILE: %s.\n"),
+                     tdir, strerror (errno)));
+      tmpfile_works = 0;
+    }
+#endif
+
+#if HAVE_DUP
+  /* If we can dup and we are creating temp files in the default location then
+     try tmpfile() + dup() + fclose() to avoid ever having a named file.  */
+  if (streq (tdir, DEFAULT_TMPDIR))
+    {
+      mode_t mask = umask (0077);
+      FILE *tfile;
+      ENULLLOOP (tfile, tmpfile ());
+      if (!tfile)
+        {
+          OS (error, NILF, "tmpfile: %s", strerror (errno));
+          return -1;
+        }
+      umask (mask);
+
+      EINTRLOOP (fd, dup (fileno (tfile)));
+      if (fd < 0)
+        OS (error, NILF, "dup: %s", strerror (errno));
+      fclose (tfile);
+    }
+#endif
+
+  return fd;
+}
